﻿using GrapeCity.ActiveReports.Aspnet.Designer;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Xml;
using System.Xml.Linq;
using System.Xml.XPath;

namespace WebDesigner_MVC
{

        /// <summary>
        /// Represents some static methods for collecting report resources.
        /// </summary>
        internal static class ReportResources
        {
            private static readonly Dictionary<string, string[]> ResourceSelectorGroups = new Dictionary<string, string[]>
                {
                        {
                                "report",
                                new []
                                {
                                        "//ns:Drillthrough/ns:ReportName",
                                        "//ns:CustomProperty[ns:Name=\"calendar:Action.Drillthrough\"]/ns:Value",
                                        "//ns:Subreport/ns:ReportName",
                                }
                        },
                        {
                                "image",
                                new []
                                {
                                        "//ns:Image[not(ns:Source) or (ns:Source = \"External\")]/ns:Value",
                                        "//ns:BackgroundImage[not(ns:Source) or (ns:Source = \"External\")]/ns:Value",
                                }
                        },
                        {
                                "theme",
                                new []
                                {
                                        "//rdl:ReportThemes/ns:ThemeUri",
                                        "//ns:CustomProperty[ns:Name = \"Theme\"]/ns:Value",
                                }
                        },
                };

            /// <summary>
            /// Collect report resource references.
            /// </summary>
            public static IEnumerable<string> GetReferences(byte[] content)
            {
                XDocument xDocument = LoadReport(content);

                if (null == xDocument.Root) return Array.Empty<string>();

                var namespaceManager = GetXmlNamespaceManager(xDocument);

                return
                        (from selectorGroup in ResourceSelectorGroups
                         from selector in selectorGroup.Value
                         from element in xDocument.XPathSelectElements(selector, namespaceManager)
                         where !string.IsNullOrEmpty(element.Value)
                         select element.Value
                         ).Distinct();
            }

            /// <summary>
            /// Update report resource references.
            /// </summary>
            /// <param name="content"></param>
            /// <param name="referencesMap"></param>
            /// <returns></returns>
            public static byte[] UpdateReferences(byte[] content, Dictionary<string, string> referencesMap)
            {
                if (!referencesMap.Any())
                {
                    return content;
                }

                var xDocument = LoadReport(content);

                var namespaceManager = GetXmlNamespaceManager(xDocument);

                foreach (var e in from selectorGroup in ResourceSelectorGroups
                                  from selector in selectorGroup.Value
                                  from element in xDocument.XPathSelectElements(selector, namespaceManager)
                                  where referencesMap.ContainsKey(element.Value)
                                  select element)
                {
                    e.SetValue(referencesMap[e.Value]);
                }

                return SaveReport(xDocument);
            }

            private static XmlNamespaceManager GetXmlNamespaceManager(XDocument document)
            {
                var defaultNamespace = document.Root.GetDefaultNamespace();
                var namespaceManager = new XmlNamespaceManager(new NameTable());
                namespaceManager.AddNamespace("ns", defaultNamespace.NamespaceName);
                namespaceManager.AddNamespace("rdl", @"http://schemas.datadynamics.com/reporting/2005/02/reportdefinition");

                return namespaceManager;
            }

            private static XDocument LoadReport(byte[] content)
            {
                try
                {
                    using (MemoryStream memoryStream = new MemoryStream(content))
                    {
                        return XDocument.Load(memoryStream);
                    }
                }
                catch (Exception exc)
                {
                    throw new InvalidReportContentException($"Report XML content is invalid. {exc.Message}");
                }
            }

            private static byte[] SaveReport(XDocument xDocument)
            {
                using (MemoryStream memoryStream = new MemoryStream())
                {
                    xDocument.Save(memoryStream);
                    memoryStream.Position = 0;
                    return memoryStream.ToArray();
                }
            }
        }
    }
